"""
LangGraph workflow export adapter.

Exports LangGraph StateGraph objects to Omium workflow format.
"""

import logging
from typing import Dict, Any, Optional, List
import inspect

logger = logging.getLogger(__name__)

try:
    from langgraph.graph import StateGraph, END, START
    from langgraph.graph.state import CompiledStateGraph
    LANGGRAPH_AVAILABLE = True
except ImportError:
    LANGGRAPH_AVAILABLE = False
    StateGraph = None
    CompiledStateGraph = None
    END = None
    START = None


def export_langgraph_workflow(
    graph: CompiledStateGraph,
    workflow_name: Optional[str] = None,
    workflow_id: Optional[str] = None
) -> Dict[str, Any]:
    """
    Export a LangGraph CompiledStateGraph to Omium workflow format.
    
    This function inspects a LangGraph graph and extracts:
    - Nodes (name, function reference)
    - Edges (from, to relationships)
    - State schema information
    
    Args:
        graph: Compiled LangGraph StateGraph instance to export
        workflow_name: Optional name for the workflow (defaults to "langgraph-workflow")
        workflow_id: Optional workflow ID (defaults to workflow_name)
        
    Returns:
        Dictionary in Omium workflow format
        
    Raises:
        ImportError: If LangGraph is not installed
        ValueError: If graph is not a valid CompiledStateGraph instance
    """
    if not LANGGRAPH_AVAILABLE:
        raise ImportError(
            "LangGraph is not installed. Install it with: pip install langgraph"
        )
    
    if not isinstance(graph, CompiledStateGraph):
        # Try to get the underlying graph if it's wrapped
        if hasattr(graph, 'graph'):
            graph = graph.graph
        else:
            raise ValueError(
                f"Expected CompiledStateGraph instance, got {type(graph)}. "
                "Note: You may need to compile the graph first with graph.compile()"
            )
    
    # Extract workflow name
    if not workflow_name:
        workflow_name = getattr(graph, 'name', None) or "langgraph-workflow"
    
    if not workflow_id:
        workflow_id = workflow_name
    
    # Extract nodes
    nodes = []
    edges = []
    
    # Try to get graph structure
    # LangGraph stores nodes and edges internally
    graph_nodes = getattr(graph, 'nodes', {})
    graph_edges = getattr(graph, 'edges', {})
    
    # If nodes/edges are not directly accessible, try alternative methods
    if not graph_nodes:
        # Try to get from _nodes attribute (internal)
        graph_nodes = getattr(graph, '_nodes', {})
    
    if not graph_edges:
        # Try to get from _edges attribute (internal)
        graph_edges = getattr(graph, '_edges', {})
    
    # Extract nodes
    if graph_nodes:
        for node_name, node_func in graph_nodes.items():
            # Skip special nodes
            if node_name in [START, END, "__start__", "__end__"]:
                continue
            
            node_dict = {
                "name": str(node_name),
                "function": _get_function_name(node_func),
            }
            
            # Try to extract function metadata
            if inspect.isfunction(node_func) or inspect.ismethod(node_func):
                if hasattr(node_func, '__name__'):
                    node_dict["function_name"] = node_func.__name__
                if hasattr(node_func, '__doc__') and node_func.__doc__:
                    node_dict["description"] = node_func.__doc__.strip().split('\n')[0]
            
            nodes.append(node_dict)
    else:
        # Fallback: try to inspect the graph structure differently
        logger.warning("Could not directly access graph nodes. Using fallback method.")
        nodes = _extract_nodes_fallback(graph)
    
    # Extract edges
    if graph_edges:
        for from_node, to_nodes in graph_edges.items():
            if not isinstance(to_nodes, list):
                to_nodes = [to_nodes]
            
            for to_node in to_nodes:
                # Handle START and END constants
                from_name = str(from_node) if from_node not in [START, "__start__"] else "START"
                to_name = str(to_node) if to_node not in [END, "__end__"] else "END"
                
                edge_dict = {
                    "from": from_name,
                    "to": to_name,
                }
                
                edges.append(edge_dict)
    else:
        # Fallback: try to reconstruct edges from graph structure
        logger.warning("Could not directly access graph edges. Using fallback method.")
        edges = _extract_edges_fallback(graph, nodes)
    
    # If we still don't have nodes/edges, create a minimal structure
    if not nodes:
        logger.warning("Could not extract nodes from graph. Creating minimal structure.")
        nodes = [{"name": "process", "function": "default_process"}]
        edges = [
            {"from": "START", "to": "process"},
            {"from": "process", "to": "END"}
        ]
    
    # Build Omium workflow definition
    workflow_definition = {
        "name": workflow_name,
        "nodes": nodes,
        "edges": edges,
    }
    
    # Build full Omium workflow format
    omium_workflow = {
        "type": "langgraph",
        "workflow_id": workflow_id,
        "agent_id": f"{workflow_id}-agent",
        "inputs": {
            "messages": [],
            "data": {
                "input": "Your input here"
            }
        },
        "definition": workflow_definition
    }
    
    logger.info(f"Exported LangGraph workflow '{workflow_name}' with {len(nodes)} nodes and {len(edges)} edges")
    
    return omium_workflow


def _get_function_name(func) -> str:
    """
    Get a string representation of a function name.
    
    Args:
        func: Function object
        
    Returns:
        Function name as string
    """
    if inspect.isfunction(func) or inspect.ismethod(func):
        return func.__name__
    elif hasattr(func, '__name__'):
        return func.__name__
    elif hasattr(func, '__class__'):
        return func.__class__.__name__
    else:
        return "unknown_function"


def _extract_nodes_fallback(graph) -> List[Dict[str, Any]]:
    """
    Fallback method to extract nodes from graph when direct access fails.
    
    Args:
        graph: LangGraph StateGraph instance
        
    Returns:
        List of node dictionaries
    """
    nodes = []
    
    # Try to use graph introspection
    if hasattr(graph, '__dict__'):
        graph_dict = graph.__dict__
        
        # Look for node-related attributes
        for key, value in graph_dict.items():
            if 'node' in key.lower() and isinstance(value, dict):
                for node_name, node_func in value.items():
                    if node_name in [START, END, "__start__", "__end__"]:
                        continue
                    
                    nodes.append({
                        "name": str(node_name),
                        "function": _get_function_name(node_func),
                    })
    
    return nodes


def _extract_edges_fallback(graph, nodes: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Fallback method to extract edges from graph when direct access fails.
    
    Args:
        graph: LangGraph StateGraph instance
        nodes: List of extracted nodes
        
    Returns:
        List of edge dictionaries
    """
    edges = []
    
    # If we have nodes, create a simple linear flow
    if nodes:
        node_names = [node["name"] for node in nodes]
        
        # Create edges: START -> first_node -> ... -> last_node -> END
        if node_names:
            # START to first node
            edges.append({
                "from": "START",
                "to": node_names[0]
            })
            
            # Between nodes
            for i in range(len(node_names) - 1):
                edges.append({
                    "from": node_names[i],
                    "to": node_names[i + 1]
                })
            
            # Last node to END
            edges.append({
                "from": node_names[-1],
                "to": "END"
            })
    else:
        # Minimal fallback
        edges = [
            {"from": "START", "to": "process"},
            {"from": "process", "to": "END"}
        ]
    
    return edges




"""
CrewAI Integration - Auto-instrumentation for CrewAI workflows

This module provides automatic instrumentation for CrewAI, capturing
traces and checkpoints for every crew execution without requiring
code changes.

Instrumentation is applied by monkey-patching key CrewAI methods:
- Crew.kickoff() - main execution method
- Crew.kickoff_async() - async execution (if available)

All agent executions and task completions are automatically traced.
"""

import logging
import functools
import asyncio
from typing import Any, Dict, List, Optional, Callable, Union
from datetime import datetime
import uuid

logger = logging.getLogger("omium.crewai")

# Store original methods for restoration
_original_kickoff = None
_original_kickoff_async = None
_original_kickoff_for_each = None
_instrumented = False


def instrument_crewai():
    """
    Instrument CrewAI for automatic tracing and checkpointing.
    
    This patches the following methods on Crew:
    - kickoff() - main synchronous execution
    - kickoff_async() - async execution (if available)
    - kickoff_for_each() - batch execution (if available)
    
    Call this function once at application startup, or use omium.init()
    which calls this automatically when CrewAI is detected.
    
    Example:
        ```python
        from omium.integrations import instrument_crewai
        instrument_crewai()
        
        # Now all CrewAI executions are automatically traced
        crew = Crew(agents=[...], tasks=[...])
        result = crew.kickoff()
        ```
    """
    global _original_kickoff, _original_kickoff_async, _original_kickoff_for_each
    global _instrumented
    
    if _instrumented:
        logger.debug("CrewAI already instrumented")
        return
    
    try:
        from crewai import Crew
    except ImportError:
        raise ImportError(
            "CrewAI is not installed. Install with: pip install crewai"
        )
    
    # Store original methods
    _original_kickoff = Crew.kickoff
    
    if hasattr(Crew, 'kickoff_async'):
        _original_kickoff_async = Crew.kickoff_async
    
    if hasattr(Crew, 'kickoff_for_each'):
        _original_kickoff_for_each = Crew.kickoff_for_each
    
    # Apply patches
    Crew.kickoff = _patched_kickoff
    
    if _original_kickoff_async:
        Crew.kickoff_async = _patched_kickoff_async
    
    if _original_kickoff_for_each:
        Crew.kickoff_for_each = _patched_kickoff_for_each
    
    _instrumented = True
    logger.info("CrewAI instrumentation applied successfully")


def uninstrument_crewai():
    """
    Remove CrewAI instrumentation.
    
    Restores the original methods on Crew.
    """
    global _original_kickoff, _original_kickoff_async, _original_kickoff_for_each
    global _instrumented
    
    if not _instrumented:
        return
    
    try:
        from crewai import Crew
    except ImportError:
        return
    
    # Restore original methods
    if _original_kickoff:
        Crew.kickoff = _original_kickoff
    if _original_kickoff_async:
        Crew.kickoff_async = _original_kickoff_async
    if _original_kickoff_for_each:
        Crew.kickoff_for_each = _original_kickoff_for_each
    
    _instrumented = False
    logger.info("CrewAI instrumentation removed")


def _get_crew_info(crew) -> Dict[str, Any]:
    """Extract information about the crew for logging."""
    info = {
        "crew_name": getattr(crew, 'name', None) or "crew",
        "agent_count": 0,
        "task_count": 0,
        "agents": [],
        "tasks": [],
    }
    
    # Get agents
    try:
        if hasattr(crew, 'agents') and crew.agents:
            info["agent_count"] = len(crew.agents)
            for agent in crew.agents[:5]:  # First 5 agents
                agent_info = {
                    "role": getattr(agent, 'role', 'unknown'),
                    "goal": str(getattr(agent, 'goal', ''))[:100],
                }
                info["agents"].append(agent_info)
    except Exception as e:
        logger.debug(f"Error extracting agent info: {e}")
    
    # Get tasks
    try:
        if hasattr(crew, 'tasks') and crew.tasks:
            info["task_count"] = len(crew.tasks)
            for task in crew.tasks[:5]:  # First 5 tasks
                task_info = {
                    "description": str(getattr(task, 'description', ''))[:100],
                }
                if hasattr(task, 'agent') and task.agent:
                    task_info["agent_role"] = getattr(task.agent, 'role', 'unknown')
                info["tasks"].append(task_info)
    except Exception as e:
        logger.debug(f"Error extracting task info: {e}")
    
    return info


def _truncate_inputs(inputs: Optional[Dict]) -> Optional[Dict]:
    """Truncate input values for logging."""
    if not inputs:
        return None
    
    result = {}
    for k, v in inputs.items():
        if isinstance(v, str) and len(v) > 200:
            result[k] = v[:200] + "..."
        elif isinstance(v, dict):
            result[k] = _truncate_inputs(v)
        else:
            result[k] = v
    return result


@functools.wraps(_original_kickoff if _original_kickoff else lambda *a, **k: None)
def _patched_kickoff(self, inputs: Optional[Dict[str, Any]] = None, **kwargs):
    """Instrumented kickoff with tracing and checkpointing."""
    from omium.integrations.core import get_current_config, is_initialized
    from omium.integrations.tracer import OmiumTracer
    
    # Skip if not initialized
    if not is_initialized():
        return _original_kickoff(self, inputs, **kwargs)
    
    omium_config = get_current_config()
    if not omium_config or not omium_config.auto_trace:
        return _original_kickoff(self, inputs, **kwargs)
    
    # Create tracer for this execution
    execution_id = str(uuid.uuid4())
    tracer = OmiumTracer(
        execution_id=execution_id,
        project=omium_config.project
    )
    
    crew_info = _get_crew_info(self)
    input_preview = _truncate_inputs(inputs)
    
    with tracer.span(
        f"crewai.kickoff:{crew_info['crew_name']}",
        span_type="crew",
        input=input_preview,
        crew_name=crew_info["crew_name"],
        agent_count=crew_info["agent_count"],
        task_count=crew_info["task_count"],
        agents=[a.get("role") for a in crew_info["agents"]],
    ) as root_span:
        try:
            logger.debug(
                f"Starting CrewAI execution: {crew_info['crew_name']} "
                f"({crew_info['agent_count']} agents, {crew_info['task_count']} tasks)"
            )
            
            # Set up step callback to capture agent activities
            original_step_callback = getattr(self, 'step_callback', None)
            
            def omium_step_callback(step_output):
                """Capture each step for tracing."""
                try:
                    tracer.add_event("crew_step", {
                        "output_preview": str(step_output)[:300] if step_output else None,
                    })
                except Exception:
                    pass
                
                # Call original callback if exists
                if original_step_callback:
                    return original_step_callback(step_output)
            
            # Temporarily set our callback
            self.step_callback = omium_step_callback
            
            try:
                result = _original_kickoff(self, inputs, **kwargs)
            finally:
                # Restore original callback
                self.step_callback = original_step_callback
            
            # Capture output
            if result is not None:
                if hasattr(result, 'raw'):
                    # CrewOutput object
                    output_str = str(result.raw)[:1000]
                else:
                    output_str = str(result)[:1000]
                root_span.set_output(output_str)
            
            logger.debug(f"CrewAI execution completed: {crew_info['crew_name']}")
            return result
            
        except Exception as e:
            root_span.set_error(e)
            logger.debug(f"CrewAI execution failed: {crew_info['crew_name']} - {e}")
            raise
            
        finally:
            tracer.flush()


@functools.wraps(_original_kickoff_async if _original_kickoff_async else lambda *a, **k: None)
async def _patched_kickoff_async(self, inputs: Optional[Dict[str, Any]] = None, **kwargs):
    """Instrumented kickoff_async with tracing and checkpointing."""
    from omium.integrations.core import get_current_config, is_initialized
    from omium.integrations.tracer import OmiumTracer
    
    if not is_initialized():
        return await _original_kickoff_async(self, inputs, **kwargs)
    
    omium_config = get_current_config()
    if not omium_config or not omium_config.auto_trace:
        return await _original_kickoff_async(self, inputs, **kwargs)
    
    execution_id = str(uuid.uuid4())
    tracer = OmiumTracer(
        execution_id=execution_id,
        project=omium_config.project
    )
    
    crew_info = _get_crew_info(self)
    input_preview = _truncate_inputs(inputs)
    
    with tracer.span(
        f"crewai.kickoff_async:{crew_info['crew_name']}",
        span_type="crew",
        input=input_preview,
        crew_name=crew_info["crew_name"],
        agent_count=crew_info["agent_count"],
        task_count=crew_info["task_count"],
        agents=[a.get("role") for a in crew_info["agents"]],
    ) as root_span:
        try:
            logger.debug(f"Starting async CrewAI execution: {crew_info['crew_name']}")
            
            result = await _original_kickoff_async(self, inputs, **kwargs)
            
            if result is not None:
                if hasattr(result, 'raw'):
                    output_str = str(result.raw)[:1000]
                else:
                    output_str = str(result)[:1000]
                root_span.set_output(output_str)
            
            logger.debug(f"Async CrewAI execution completed: {crew_info['crew_name']}")
            return result
            
        except Exception as e:
            root_span.set_error(e)
            logger.debug(f"Async CrewAI execution failed: {crew_info['crew_name']} - {e}")
            raise
            
        finally:
            await tracer.aflush()


def _patched_kickoff_for_each(self, inputs: List[Dict[str, Any]], **kwargs):
    """Instrumented kickoff_for_each with tracing."""
    from omium.integrations.core import get_current_config, is_initialized
    from omium.integrations.tracer import OmiumTracer
    
    if not is_initialized():
        return _original_kickoff_for_each(self, inputs, **kwargs)
    
    omium_config = get_current_config()
    if not omium_config or not omium_config.auto_trace:
        return _original_kickoff_for_each(self, inputs, **kwargs)
    
    execution_id = str(uuid.uuid4())
    tracer = OmiumTracer(
        execution_id=execution_id,
        project=omium_config.project
    )
    
    crew_info = _get_crew_info(self)
    
    with tracer.span(
        f"crewai.kickoff_for_each:{crew_info['crew_name']}",
        span_type="crew_batch",
        crew_name=crew_info["crew_name"],
        batch_size=len(inputs),
        agent_count=crew_info["agent_count"],
        task_count=crew_info["task_count"],
    ) as root_span:
        try:
            logger.debug(
                f"Starting CrewAI batch execution: {crew_info['crew_name']} "
                f"({len(inputs)} inputs)"
            )
            
            results = _original_kickoff_for_each(self, inputs, **kwargs)
            
            root_span.set_output(f"Processed {len(inputs)} inputs, got {len(results)} results")
            logger.debug(f"CrewAI batch execution completed: {crew_info['crew_name']}")
            
            return results
            
        except Exception as e:
            root_span.set_error(e)
            raise
            
        finally:
            tracer.flush()


def is_instrumented() -> bool:
    """Check if CrewAI is currently instrumented."""
    return _instrumented



"""
Omium CLI Streaming - Live log streaming and real-time monitoring.

This module provides utilities for streaming logs and monitoring executions in real-time.
"""

import asyncio
import json
from datetime import datetime
from typing import Optional, Callable, AsyncIterator

import httpx
import websockets
from websockets.exceptions import ConnectionClosed

from omium.output import (
    console, print_success, print_error, print_warning, print_info,
    print_header, print_divider, OmiumSpinner
)


async def stream_execution_logs(
    execution_id: str,
    execution_engine_url: str = "http://localhost:8000",
    websocket_url: Optional[str] = None,
    follow: bool = True,
    tail: int = 50,
    on_log: Optional[Callable[[dict], None]] = None
) -> AsyncIterator[dict]:
    """
    Stream logs from an execution in real-time.
    
    Args:
        execution_id: The execution ID to stream logs from
        execution_engine_url: Base URL of the execution engine
        websocket_url: Optional WebSocket URL for real-time streaming
        follow: If True, continue streaming new logs as they arrive
        tail: Number of recent logs to fetch initially
        on_log: Optional callback for each log entry
        
    Yields:
        Log entry dictionaries
    """
    # First, fetch existing logs via HTTP
    async with httpx.AsyncClient() as client:
        try:
            response = await client.get(
                f"{execution_engine_url}/api/v1/executions/{execution_id}/logs",
                params={"tail": tail}
            )
            
            if response.status_code == 200:
                data = response.json()
                logs = data.get("logs", [])
                
                for log in logs:
                    if on_log:
                        on_log(log)
                    yield log
                    
            elif response.status_code == 404:
                print_error(f"Execution not found: {execution_id}")
                return
            else:
                print_warning(f"Could not fetch logs: {response.status_code}")
                
        except httpx.RequestError as e:
            print_warning(f"Could not connect to fetch logs: {e}")
    
    # If follow mode, try WebSocket streaming
    if follow:
        ws_url = websocket_url or execution_engine_url.replace("http", "ws")
        ws_endpoint = f"{ws_url}/ws/executions/{execution_id}/logs"
        
        try:
            async with websockets.connect(ws_endpoint) as ws:
                console.print("[dim]Connected to live log stream...[/dim]")
                while True:
                    try:
                        message = await ws.recv()
                        log = json.loads(message)
                        if on_log:
                            on_log(log)
                        yield log
                    except ConnectionClosed:
                        console.print("[dim]Log stream disconnected[/dim]")
                        break
        except Exception as e:
            # WebSocket not available, fall back to polling
            console.print(f"[dim]WebSocket not available, using polling mode...[/dim]")
            await _poll_logs(execution_id, execution_engine_url, on_log)


async def _poll_logs(
    execution_id: str,
    execution_engine_url: str,
    on_log: Optional[Callable[[dict], None]] = None,
    poll_interval: float = 2.0
):
    """Poll for new logs when WebSocket is not available."""
    seen_log_ids = set()
    
    async with httpx.AsyncClient() as client:
        while True:
            try:
                response = await client.get(
                    f"{execution_engine_url}/api/v1/executions/{execution_id}/logs"
                )
                
                if response.status_code == 200:
                    data = response.json()
                    logs = data.get("logs", [])
                    
                    for log in logs:
                        log_id = log.get("id") or log.get("timestamp")
                        if log_id and log_id not in seen_log_ids:
                            seen_log_ids.add(log_id)
                            if on_log:
                                on_log(log)
                    
                    # Check if execution is complete
                    exec_response = await client.get(
                        f"{execution_engine_url}/api/v1/executions/{execution_id}"
                    )
                    if exec_response.status_code == 200:
                        execution = exec_response.json()
                        status = execution.get("status", "")
                        if status in ("completed", "failed", "cancelled"):
                            console.print(f"[dim]Execution {status}. Stopping log stream.[/dim]")
                            break
                            
            except httpx.RequestError:
                pass  # Ignore connection errors during polling
                
            await asyncio.sleep(poll_interval)


def format_log_entry(log: dict) -> str:
    """Format a log entry for display."""
    timestamp = log.get("timestamp", "")
    level = log.get("level", "INFO").upper()
    message = log.get("message", "")
    agent_id = log.get("agent_id", "")
    
    # Format timestamp
    if timestamp:
        try:
            dt = datetime.fromisoformat(timestamp.replace("Z", "+00:00"))
            timestamp = dt.strftime("%H:%M:%S.%f")[:-3]
        except:
            timestamp = timestamp[:12]
    
    # Color based on level
    level_colors = {
        "DEBUG": "dim",
        "INFO": "blue",
        "WARNING": "yellow",
        "ERROR": "red",
        "CRITICAL": "bold red"
    }
    level_color = level_colors.get(level, "white")
    
    # Format the line
    parts = []
    if timestamp:
        parts.append(f"[dim]{timestamp}[/dim]")
    parts.append(f"[{level_color}]{level:7}[/{level_color}]")
    if agent_id:
        parts.append(f"[cyan][{agent_id}][/cyan]")
    parts.append(message)
    
    return " ".join(parts)


def print_log_entry(log: dict) -> None:
    """Print a formatted log entry to the console."""
    formatted = format_log_entry(log)
    console.print(formatted)


async def get_execution_status(
    execution_id: str,
    execution_engine_url: str = "http://localhost:8000"
) -> Optional[dict]:
    """Get current execution status."""
    async with httpx.AsyncClient() as client:
        try:
            response = await client.get(
                f"{execution_engine_url}/api/v1/executions/{execution_id}"
            )
            if response.status_code == 200:
                return response.json()
        except httpx.RequestError:
            pass
    return None


async def watch_execution(
    execution_id: str,
    execution_engine_url: str = "http://localhost:8000",
    update_callback: Optional[Callable[[dict], None]] = None,
    poll_interval: float = 1.0
) -> None:
    """
    Watch an execution for status changes.
    
    Args:
        execution_id: The execution ID to watch
        execution_engine_url: Base URL of the execution engine
        update_callback: Called when status changes
        poll_interval: Seconds between status checks
    """
    last_status = None
    
    async with httpx.AsyncClient() as client:
        while True:
            try:
                response = await client.get(
                    f"{execution_engine_url}/api/v1/executions/{execution_id}"
                )
                
                if response.status_code == 200:
                    execution = response.json()
                    current_status = execution.get("status")
                    
                    if current_status != last_status:
                        last_status = current_status
                        if update_callback:
                            update_callback(execution)
                    
                    # Stop watching if execution is done
                    if current_status in ("completed", "failed", "cancelled"):
                        break
                        
            except httpx.RequestError:
                pass
                
            await asyncio.sleep(poll_interval)



"""
Checkpoint decorator and context manager for Omium SDK.
"""

import asyncio
import functools
import inspect
import json
import logging
from typing import Callable, Any, Optional, List, Dict, Union
from contextlib import asynccontextmanager

from .client import OmiumClient, get_client, CheckpointError, CheckpointValidationError
from .remote_client import RemoteOmiumClient
from .config import get_config

logger = logging.getLogger(__name__)


def _serialize_state(obj: Any) -> Dict[str, Any]:
    """
    Serialize object to dictionary for checkpoint storage.
    
    Handles:
    - Dicts, lists, primitives
    - Objects with __dict__
    - JSON-serializable types
    """
    if isinstance(obj, dict):
        return {k: _serialize_state(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple)):
        return [_serialize_state(item) for item in obj]
    elif isinstance(obj, (str, int, float, bool, type(None))):
        return obj
    elif hasattr(obj, "__dict__"):
        return _serialize_state(obj.__dict__)
    else:
        # Try JSON serialization
        try:
            return json.loads(json.dumps(obj, default=str))
        except (TypeError, ValueError):
            return {"_type": type(obj).__name__, "_repr": str(obj)}


def _capture_state(func: Callable, args: tuple, kwargs: dict, result: Any = None) -> Dict[str, Any]:
    """
    Capture function state for checkpointing.
    
    Args:
        func: Function being checkpointed
        args: Function arguments
        kwargs: Function keyword arguments
        result: Function result (if post-execution)
    
    Returns:
        State dictionary
    """
    sig = inspect.signature(func)
    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()
    
    state = {
        "function": func.__name__,
        "module": func.__module__,
        "args": {},
        "kwargs": {},
    }
    
    # Capture arguments
    for param_name, param_value in bound_args.arguments.items():
        if param_name in sig.parameters:
            param = sig.parameters[param_name]
            if param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                state["args"][param_name] = _serialize_state(param_value)
            elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                state["kwargs"][param_name] = _serialize_state(param_value)
    
    # Capture result if provided
    if result is not None:
        state["result"] = _serialize_state(result)
    
    return state


def checkpoint(
    name: str,
    preconditions: Optional[List[str]] = None,
    postconditions: Optional[List[str]] = None,
    client: Optional[Union[OmiumClient, RemoteOmiumClient]] = None,
    execution_id: Optional[str] = None,
    agent_id: Optional[str] = None,
):
    """
    Decorator to create a checkpoint before/after function execution.
    
    Args:
        name: Checkpoint name (must be unique per execution)
        preconditions: List of precondition strings (e.g., ["amount > 0", "user exists"])
        postconditions: List of postcondition strings
        client: Optional OmiumClient instance (uses global if not provided)
        execution_id: Optional execution ID (uses client context if not provided)
        agent_id: Optional agent ID (uses client context if not provided)
    
    Usage:
        @checkpoint("validate_data", preconditions=["data is not None"])
        async def validate(data):
            return processed_data
    
    The decorator will:
    1. Capture function arguments (pre-execution state)
    2. Execute function
    3. Capture return value (post-execution state)
    4. Create checkpoint with combined state
    5. Validate postconditions
    """
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            # Get client (use provided or create based on config)
            if client is None:
                config = get_config()
                if config.use_remote_api:
                    # Use remote HTTP client
                    checkpoint_client = RemoteOmiumClient()
                else:
                    # Use local gRPC client
                    checkpoint_client = get_client()
                    # Ensure client is connected
                    if hasattr(checkpoint_client, '_checkpoint_stub') and checkpoint_client._checkpoint_stub is None:
                        await checkpoint_client.connect()
            else:
                checkpoint_client = client
                # Ensure gRPC client is connected if needed
                if isinstance(checkpoint_client, OmiumClient):
                    if checkpoint_client._checkpoint_stub is None:
                        await checkpoint_client.connect()
            
            # Capture pre-execution state
            pre_state = _capture_state(func, args, kwargs)
            
            try:
                # Execute function
                result = await func(*args, **kwargs)
                
                # Capture post-execution state
                post_state = _capture_state(func, args, kwargs, result)
                
                # Create checkpoint with combined state
                checkpoint_state = {
                    "pre_execution": pre_state,
                    "post_execution": post_state,
                    "function": func.__name__,
                    "module": func.__module__,
                }
                
                checkpoint_id = await checkpoint_client.create_checkpoint(
                    checkpoint_name=name,
                    state=checkpoint_state,
                    execution_id=execution_id,
                    agent_id=agent_id,
                    preconditions=preconditions,
                    postconditions=postconditions,
                    metadata={
                        "function": func.__name__,
                        "module": func.__module__,
                        "decorator": "checkpoint",
                    },
                )
                
                logger.debug(f"Checkpoint created: {checkpoint_id} for {name}")
                return result
                
            except CheckpointValidationError as e:
                logger.error(f"Checkpoint validation failed for {name}: {e}")
                raise
            except Exception as e:
                logger.error(f"Error in checkpointed function {name}: {e}")
                # On failure, we could rollback, but for now just re-raise
                raise
        
        @functools.wraps(func)
        def sync_wrapper(*args, **kwargs):
            # For sync functions, run in event loop
            loop = asyncio.get_event_loop()
            if loop.is_running():
                # If loop is running, create task
                return asyncio.create_task(async_wrapper(*args, **kwargs))
            else:
                return loop.run_until_complete(async_wrapper(*args, **kwargs))
        
        # Determine if function is async
        if inspect.iscoroutinefunction(func):
            return async_wrapper
        else:
            return sync_wrapper
    
    return decorator


class Checkpoint:
    """
    Context manager for manual checkpoint control.
    
    Usage:
        async with Checkpoint("important_state") as cp:
            # Critical code here
            result = await do_critical_thing()
            # Checkpoint saved automatically on exit
    """
    
    def __init__(
        self,
        name: str,
        client: Optional[Union[OmiumClient, RemoteOmiumClient]] = None,
        execution_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        preconditions: Optional[List[str]] = None,
        postconditions: Optional[List[str]] = None,
    ):
        """
        Initialize checkpoint context manager.
        
        Args:
            name: Checkpoint name
            client: Optional OmiumClient instance
            execution_id: Optional execution ID
            agent_id: Optional agent ID
            preconditions: Optional precondition list
            postconditions: Optional postcondition list
        """
        self.name = name
        # Get client based on config if not provided
        if client is None:
            config = get_config()
            if config.use_remote_api:
                self.client = RemoteOmiumClient()
            else:
                self.client = get_client()
        else:
            self.client = client
        self.execution_id = execution_id
        self.agent_id = agent_id
        self.preconditions = preconditions
        self.postconditions = postconditions
        self.checkpoint_id: Optional[str] = None
        self._state: Dict[str, Any] = {}
    
    async def __aenter__(self):
        """Enter checkpoint context."""
        # Ensure gRPC client is connected if needed
        if isinstance(self.client, OmiumClient):
            if self.client._checkpoint_stub is None:
                await self.client.connect()
        
        # Capture initial state
        self._state = {
            "checkpoint_name": self.name,
            "entered_at": asyncio.get_event_loop().time(),
        }
        
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Exit checkpoint context."""
        # Update state with exit information
        self._state.update({
            "exited_at": asyncio.get_event_loop().time(),
            "exception": str(exc_val) if exc_val else None,
            "exception_type": exc_type.__name__ if exc_type else None,
        })
        
        # Create checkpoint
        try:
            self.checkpoint_id = await self.client.create_checkpoint(
                checkpoint_name=self.name,
                state=self._state,
                execution_id=self.execution_id,
                agent_id=self.agent_id,
                preconditions=self.preconditions,
                postconditions=self.postconditions,
                metadata={
                    "context_manager": True,
                    "exception": str(exc_val) if exc_val else None,
                },
            )
            logger.debug(f"Checkpoint created: {self.checkpoint_id} for {self.name}")
        except Exception as e:
            logger.error(f"Failed to create checkpoint {self.name}: {e}")
            # Don't raise - let original exception propagate if any
        
        # If exception occurred, could rollback here
        # For now, just log it
        if exc_type is not None:
            logger.warning(f"Exception in checkpoint context {self.name}: {exc_val}")
        
        return False  # Don't suppress exceptions
    
    def update_state(self, **kwargs):
        """
        Update checkpoint state (call within context).
        
        Args:
            **kwargs: State updates
        """
        self._state.update(kwargs)



"""
Consensus client for multi-agent coordination.
"""

import asyncio
import json
import logging
from typing import Optional, Dict, Any, List

import grpc
from google.protobuf import struct_pb2
from google.protobuf import json_format

# Import generated proto code
try:
    from omium.proto.consensus import consensus_pb2, consensus_pb2_grpc
except ImportError:
    try:
        from omium.proto import consensus_pb2, consensus_pb2_grpc
    except ImportError:
        consensus_pb2 = None
        consensus_pb2_grpc = None

logger = logging.getLogger(__name__)


class ConsensusError(Exception):
    """Base exception for consensus operations."""
    pass


class ConsensusClient:
    """
    Async gRPC client for Consensus Coordinator service.
    
    Provides multi-agent consensus and handoff capabilities.
    """
    
    def __init__(
        self,
        consensus_coordinator_url: str = "localhost:7002",
        timeout: float = 30.0,
    ):
        """
        Initialize consensus client.
        
        Args:
            consensus_coordinator_url: Consensus Coordinator gRPC endpoint
            timeout: Request timeout in seconds
        """
        self.consensus_coordinator_url = consensus_coordinator_url
        self.timeout = timeout
        self._channel: Optional[grpc.aio.Channel] = None
        self._consensus_stub: Optional[Any] = None
        self._consensus_pb2 = None
        self._consensus_pb2_grpc = None
    
    async def connect(self):
        """Create gRPC channel and stubs."""
        # Try multiple import paths
        try:
            from omium.proto.consensus import consensus_pb2, consensus_pb2_grpc
            self._consensus_pb2 = consensus_pb2
            self._consensus_pb2_grpc = consensus_pb2_grpc
        except ImportError as e1:
            try:
                from omium.proto import consensus_pb2, consensus_pb2_grpc
                self._consensus_pb2 = consensus_pb2
                self._consensus_pb2_grpc = consensus_pb2_grpc
            except ImportError as e2:
                raise ImportError(
                    f"Consensus proto code not found. Tried: omium.proto.consensus (error: {e1}), "
                    f"omium.proto (error: {e2}). "
                    "Ensure proto files are generated and included in package."
                )
        
        self._channel = grpc.aio.insecure_channel(self.consensus_coordinator_url)
        self._consensus_stub = self._consensus_pb2_grpc.ConsensusServiceStub(self._channel)
        
        # Test connection
        try:
            await asyncio.wait_for(self._channel.channel_ready(), timeout=5.0)
            logger.info(f"Connected to consensus-coordinator at {self.consensus_coordinator_url}")
        except asyncio.TimeoutError:
            raise ConnectionError(f"Failed to connect to {self.consensus_coordinator_url}")
    
    async def close(self):
        """Close gRPC channel."""
        if self._channel:
            await self._channel.close()
            logger.info("Closed consensus client connection")
    
    def _dict_to_struct(self, data: Dict[str, Any]) -> struct_pb2.Struct:
        """Convert Python dict to protobuf Struct."""
        return json_format.ParseDict(data, struct_pb2.Struct())
    
    def _struct_to_dict(self, struct: struct_pb2.Struct) -> Dict[str, Any]:
        """Convert protobuf Struct to Python dict."""
        if struct is None:
            return {}
        return json_format.MessageToDict(struct)
    
    async def broadcast_message(
        self,
        execution_id: str,
        sender_agent_id: str,
        receiver_agent_ids: List[str],
        message_type: int,
        payload: Dict[str, Any],
        acks_required: Optional[int] = None,
        timeout_seconds: int = 5,
    ) -> Dict[str, Any]:
        """
        Broadcast message to agents with consensus guarantees.
        
        Args:
            execution_id: Execution ID
            sender_agent_id: Sender agent ID
            receiver_agent_ids: List of receiver agent IDs
            message_type: Message type (1=HANDOFF, 2=STATE_UPDATE, etc.)
            payload: Message payload
            acks_required: Number of acknowledgments required (default: majority)
            timeout_seconds: Timeout for consensus
        
        Returns:
            Dictionary with consensus result
        """
        if self._consensus_stub is None:
            raise RuntimeError("Client not connected. Call connect() first.")
        
        if acks_required is None:
            # Default to majority
            acks_required = (len(receiver_agent_ids) // 2) + 1
        
        request = self._consensus_pb2.BroadcastMessageRequest(
            execution_id=execution_id,
            sender_agent_id=sender_agent_id,
            receiver_agent_ids=receiver_agent_ids,
            message_type=message_type,
            payload=self._dict_to_struct(payload),
            acks_required=acks_required,
            timeout_seconds=timeout_seconds,
        )
        
        try:
            response = await asyncio.wait_for(
                self._consensus_stub.BroadcastMessage(request),
                timeout=self.timeout
            )
            
            if not response.success:
                raise ConsensusError(response.error_message or "Failed to broadcast message")
            
            result = {
                "message_id": response.message.id if response.message else None,
                "consensus_reached": response.consensus_reached,
                "acks_received": response.acks_received,
                "acks_required": response.acks_required,
                "acknowledged_by": list(response.acknowledged_by),
            }
            
            if response.message and response.message.payload:
                result["payload"] = self._struct_to_dict(response.message.payload)
            
            return result
            
        except grpc.RpcError as e:
            raise ConsensusError(f"gRPC error: {e.details()}")
        except asyncio.TimeoutError:
            raise ConsensusError(f"Request timeout after {self.timeout}s")
    
    async def acknowledge_message(
        self,
        message_id: str,
        agent_id: str,
        valid: bool = True,
        error_message: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Acknowledge receipt of a consensus message.
        
        Args:
            message_id: Message ID to acknowledge
            agent_id: Agent ID acknowledging
            valid: Whether message is valid
            error_message: Error message if invalid
        
        Returns:
            Dictionary with acknowledgment result
        """
        if self._consensus_stub is None:
            raise RuntimeError("Client not connected. Call connect() first.")
        
        request = self._consensus_pb2.AcknowledgeMessageRequest(
            message_id=message_id,
            agent_id=agent_id,
            valid=valid,
            error_message=error_message or "",
        )
        
        try:
            response = await asyncio.wait_for(
                self._consensus_stub.AcknowledgeMessage(request),
                timeout=self.timeout
            )
            
            return {
                "success": response.success,
                "consensus_reached": response.consensus_reached,
            }
            
        except grpc.RpcError as e:
            raise ConsensusError(f"gRPC error: {e.details()}")
        except asyncio.TimeoutError:
            raise ConsensusError(f"Request timeout after {self.timeout}s")
    
    async def propose_handoff(
        self,
        execution_id: str,
        from_agent_id: str,
        to_agent_id: str,
        state: Dict[str, Any],
        context: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Propose handoff from one agent to another.
        
        Args:
            execution_id: Execution ID
            from_agent_id: Source agent ID
            to_agent_id: Target agent ID
            state: State to hand off
            context: Optional context
        
        Returns:
            Dictionary with handoff proposal details
        """
        if self._consensus_stub is None:
            raise RuntimeError("Client not connected. Call connect() first.")
        
        request = self._consensus_pb2.ProposeHandoffRequest(
            execution_id=execution_id,
            from_agent_id=from_agent_id,
            to_agent_id=to_agent_id,
            state=self._dict_to_struct(state),
            context=self._dict_to_struct(context) if context else None,
        )
        
        try:
            response = await asyncio.wait_for(
                self._consensus_stub.ProposeHandoff(request),
                timeout=self.timeout
            )
            
            if not response.success:
                raise ConsensusError(response.error_message or "Failed to propose handoff")
            
            proposal = response.proposal
            result = {
                "proposal_id": proposal.id if proposal else None,
                "execution_id": proposal.execution_id if proposal else None,
                "from_agent_id": proposal.from_agent_id if proposal else None,
                "to_agent_id": proposal.to_agent_id if proposal else None,
                "success": response.success,
            }
            
            if proposal and proposal.state:
                result["state"] = self._struct_to_dict(proposal.state)
            if proposal and proposal.context:
                result["context"] = self._struct_to_dict(proposal.context)
            
            return result
            
        except grpc.RpcError as e:
            raise ConsensusError(f"gRPC error: {e.details()}")
        except asyncio.TimeoutError:
            raise ConsensusError(f"Request timeout after {self.timeout}s")
    
    async def vote_on_handoff(
        self,
        proposal_id: str,
        agent_id: str,
        vote: bool,
        reason: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Vote on a handoff proposal.
        
        Args:
            proposal_id: Handoff proposal ID
            agent_id: Agent ID voting
            vote: True to accept, False to reject
            reason: Reason for vote
        
        Returns:
            Dictionary with voting result
        """
        if self._consensus_stub is None:
            raise RuntimeError("Client not connected. Call connect() first.")
        
        request = self._consensus_pb2.VoteOnHandoffRequest(
            proposal_id=proposal_id,
            agent_id=agent_id,
            vote=vote,
            reason=reason or "",
        )
        
        try:
            response = await asyncio.wait_for(
                self._consensus_stub.VoteOnHandoff(request),
                timeout=self.timeout
            )
            
            if not response.success:
                raise ConsensusError(response.error_message or "Failed to vote")
            
            return {
                "success": response.success,
                "proposal_accepted": response.proposal_accepted,
                "votes_for": response.votes_for,
                "votes_against": response.votes_against,
                "votes_required": response.votes_required,
            }
            
        except grpc.RpcError as e:
            raise ConsensusError(f"gRPC error: {e.details()}")
        except asyncio.TimeoutError:
            raise ConsensusError(f"Request timeout after {self.timeout}s")
    
    async def get_consensus_status(
        self,
        execution_id: str,
        message_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Get consensus status for a message or execution.
        
        Args:
            execution_id: Execution ID
            message_id: Optional message ID
        
        Returns:
            Dictionary with consensus status
        """
        if self._consensus_stub is None:
            raise RuntimeError("Client not connected. Call connect() first.")
        
        request = self._consensus_pb2.GetConsensusStatusRequest(
            execution_id=execution_id,
            message_id=message_id or "",
        )
        
        try:
            response = await asyncio.wait_for(
                self._consensus_stub.GetConsensusStatus(request),
                timeout=self.timeout
            )
            
            return {
                "consensus_reached": response.consensus_reached,
                "acks_received": response.acks_received,
                "acks_required": response.acks_required,
                "acknowledged_by": list(response.acknowledged_by),
                "pending_agents": list(response.pending_agents),
            }
            
        except grpc.RpcError as e:
            raise ConsensusError(f"gRPC error: {e.details()}")
        except asyncio.TimeoutError:
            raise ConsensusError(f"Request timeout after {self.timeout}s")




"""
Omium framework adapters for exporting workflows from external frameworks.
"""

from .crewai_adapter import export_crewai_workflow
from .langgraph_adapter import export_langgraph_workflow

__all__ = [
    "export_crewai_workflow",
    "export_langgraph_workflow",
]




"""
Configuration management for Omium SDK.
"""

import json
import os
import logging
from pathlib import Path
from typing import Optional, Dict, Any
from pydantic import BaseModel, Field, field_validator

logger = logging.getLogger(__name__)

# Default config directory
CONFIG_DIR = Path.home() / ".omium"
CONFIG_FILE = CONFIG_DIR / "config.json"


class OmiumConfig(BaseModel):
    """Omium SDK configuration model."""
    
    api_key: Optional[str] = Field(None, description="API key for authentication")
    api_url: str = Field(
        default="https://api.omium.ai",
        description="Omium API base URL"
    )
    region: str = Field(
        default="us-east-1",
        description="AWS region (us-east-1, us-west-2, eu-west-1, etc.)"
    )
    checkpoint_manager_url: Optional[str] = Field(
        None,
        description="Local checkpoint manager URL (for local development)"
    )
    use_remote_api: bool = Field(
        default=True,
        description="Whether to use remote API (True) or local gRPC (False)"
    )
    timeout: float = Field(
        default=30.0,
        description="Request timeout in seconds"
    )
    retry_attempts: int = Field(
        default=3,
        description="Number of retry attempts for failed requests"
    )
    
    @field_validator("api_key")
    @classmethod
    def validate_api_key(cls, v: Optional[str]) -> Optional[str]:
        """Validate API key format."""
        if v is None:
            return v
        if not v.startswith("omium_"):
            raise ValueError("API key must start with 'omium_'")
        if len(v) < 20:
            raise ValueError("API key appears to be invalid (too short)")
        return v
    
    @field_validator("region")
    @classmethod
    def validate_region(cls, v: str) -> str:
        """Validate region format."""
        # Basic validation - can be expanded
        if not v or len(v) < 3:
            raise ValueError("Region must be at least 3 characters")
        return v.lower()
    
    class Config:
        """Pydantic config."""
        extra = "forbid"
        json_encoders = {
            Path: str,
        }


class ConfigManager:
    """Manages Omium SDK configuration."""
    
    def __init__(self, config_file: Optional[Path] = None):
        """
        Initialize config manager.
        
        Args:
            config_file: Optional path to config file (defaults to ~/.omium/config.json)
        """
        self.config_file = config_file or CONFIG_FILE
        self._config: Optional[OmiumConfig] = None
    
    def load(self) -> OmiumConfig:
        """
        Load configuration from file and environment variables.
        
        Priority:
        1. Environment variables (OMIUM_API_KEY, OMIUM_API_URL, etc.)
        2. Config file (~/.omium/config.json)
        3. Default values
        
        Returns:
            OmiumConfig instance
        """
        # Start with defaults
        config_dict: Dict[str, Any] = {}
        
        # Load from config file if it exists
        if self.config_file.exists():
            try:
                with open(self.config_file, "r") as f:
                    file_config = json.load(f)
                    config_dict.update(file_config)
                logger.debug(f"Loaded config from {self.config_file}")
            except Exception as e:
                logger.warning(f"Failed to load config file {self.config_file}: {e}")
        
        # Override with environment variables
        env_mapping = {
            "OMIUM_API_KEY": "api_key",
            "OMIUM_API_URL": "api_url",
            "OMIUM_REGION": "region",
            "OMIUM_CHECKPOINT_MANAGER_URL": "checkpoint_manager_url",
            "OMIUM_USE_REMOTE_API": "use_remote_api",
            "OMIUM_TIMEOUT": "timeout",
            "OMIUM_RETRY_ATTEMPTS": "retry_attempts",
        }
        
        for env_var, config_key in env_mapping.items():
            env_value = os.getenv(env_var)
            if env_value is not None:
                # Type conversion
                if config_key in ["use_remote_api"]:
                    config_dict[config_key] = env_value.lower() in ("true", "1", "yes")
                elif config_key in ["timeout"]:
                    config_dict[config_key] = float(env_value)
                elif config_key in ["retry_attempts"]:
                    config_dict[config_key] = int(env_value)
                else:
                    config_dict[config_key] = env_value
        
        # Create config object
        try:
            self._config = OmiumConfig(**config_dict)
        except Exception as e:
            logger.error(f"Invalid configuration: {e}")
            # Return default config on error
            self._config = OmiumConfig()
        
        return self._config
    
    def save(self, config: Optional[OmiumConfig] = None) -> None:
        """
        Save configuration to file.
        
        Args:
            config: Optional config to save (uses current config if not provided)
        """
        if config is None:
            config = self._config or self.load()
        
        # Create config directory if it doesn't exist
        self.config_file.parent.mkdir(parents=True, exist_ok=True)
        
        # Prepare config dict (exclude None values for optional fields)
        config_dict = config.model_dump(exclude_none=True)
        
        # Don't save API key to file if it's set (security)
        # User should use environment variable instead
        if "api_key" in config_dict:
            logger.warning(
                "API key found in config. Consider using OMIUM_API_KEY environment variable instead."
            )
        
        try:
            with open(self.config_file, "w") as f:
                json.dump(config_dict, f, indent=2)
            logger.info(f"Saved config to {self.config_file}")
        except Exception as e:
            logger.error(f"Failed to save config to {self.config_file}: {e}")
            raise
    
    def get(self) -> OmiumConfig:
        """Get current configuration (loads if not already loaded)."""
        if self._config is None:
            self._config = self.load()
        return self._config
    
    def update(self, **kwargs) -> OmiumConfig:
        """
        Update configuration values.
        
        Args:
            **kwargs: Configuration values to update
            
        Returns:
            Updated OmiumConfig
        """
        current = self.get()
        updated_dict = current.model_dump()
        updated_dict.update(kwargs)
        self._config = OmiumConfig(**updated_dict)
        return self._config
    
    def validate_api_key(self, api_key: str) -> bool:
        """
        Validate API key format.
        
        Args:
            api_key: API key to validate
            
        Returns:
            True if format is valid
        """
        try:
            OmiumConfig(api_key=api_key)
            return True
        except Exception:
            return False


# Global config manager instance
_config_manager: Optional[ConfigManager] = None


def get_config_manager() -> ConfigManager:
    """Get or create global config manager instance."""
    global _config_manager
    if _config_manager is None:
        _config_manager = ConfigManager()
    return _config_manager


def get_config() -> OmiumConfig:
    """Get current configuration."""
    return get_config_manager().get()




"""
Unit tests for Omium SDK checkpoint functionality.
"""

import pytest
import asyncio
from types import SimpleNamespace
from unittest.mock import AsyncMock, MagicMock, patch
from omium import checkpoint, Checkpoint, OmiumClient
from omium.client import CheckpointError, CheckpointNotFoundError


@pytest.fixture
def mock_client():
    """Create a mock OmiumClient."""
    client = OmiumClient(checkpoint_manager_url="localhost:7001")
    client._checkpoint_stub = AsyncMock()
    client._channel = MagicMock()
    # Provide lightweight protobuf request factories so tests do not rely on compiled protos
    def _dummy_request(**kwargs):
        return SimpleNamespace(**kwargs)

    client._checkpoint_pb2 = SimpleNamespace(
        CreateCheckpointRequest=_dummy_request,
        GetCheckpointRequest=_dummy_request,
        ListCheckpointsRequest=_dummy_request,
        RollbackToCheckpointRequest=_dummy_request,
    )
    return client


@pytest.mark.asyncio
async def test_checkpoint_decorator_success(mock_client):
    """Test @checkpoint decorator with successful execution."""
    from omium.client import set_client
    set_client(mock_client)
    
    # Mock successful checkpoint creation
    mock_response = MagicMock()
    mock_response.success = True
    mock_response.checkpoint.id = "cp_123"
    mock_response.error_message = None
    mock_client.create_checkpoint = AsyncMock(return_value="cp_123")
    
    # The new @checkpoint decorator doesn't take a client argument
    # It uses the global client set via set_client()
    @checkpoint("test_checkpoint")
    async def test_function(data):
        return {"result": data}
    
    result = await test_function({"value": 42})
    
    assert result == {"result": {"value": 42}}
    # Note: The new decorator may not call create_checkpoint in the same way
    # This test verifies the function executes correctly with the decorator


@pytest.mark.asyncio
async def test_checkpoint_context_manager(mock_client):
    """Test Checkpoint context manager."""
    mock_client.create_checkpoint = AsyncMock(return_value="cp_456")
    
    async with Checkpoint("test_context", client=mock_client) as cp:
        cp.update_state(step="test")
    
    assert mock_client.create_checkpoint.called
    assert cp.checkpoint_id == "cp_456"


@pytest.mark.asyncio
async def test_client_create_checkpoint(mock_client):
    """Test client checkpoint creation."""
    # Mock gRPC response
    mock_response = MagicMock()
    mock_response.success = True
    mock_response.checkpoint.id = "cp_789"
    mock_response.error_message = None
    
    mock_client._checkpoint_stub.CreateCheckpoint = AsyncMock(return_value=mock_response)
    
    checkpoint_id = await mock_client.create_checkpoint(
        checkpoint_name="test",
        state={"data": "test"},
        execution_id="exec_1",
    )
    
    assert checkpoint_id == "cp_789"
    assert mock_client._checkpoint_stub.CreateCheckpoint.called


@pytest.mark.asyncio
async def test_client_get_checkpoint(mock_client):
    """Test client checkpoint retrieval."""
    from google.protobuf import timestamp_pb2, struct_pb2
    
    # Mock response
    mock_checkpoint = MagicMock()
    mock_checkpoint.id = "cp_123"
    mock_checkpoint.execution_id = "exec_1"
    mock_checkpoint.checkpoint_name = "test"
    mock_checkpoint.state_size_bytes = 100
    mock_checkpoint.checksum = "abc123"
    mock_checkpoint.preconditions = []
    mock_checkpoint.postconditions = []
    mock_checkpoint.validation_passed = True
    mock_checkpoint.compression_type = "none"
    mock_checkpoint.created_at = timestamp_pb2.Timestamp()
    
    state_struct = struct_pb2.Struct()
    state_struct.update({"data": "test"})

    mock_response = MagicMock()
    mock_response.success = True
    mock_response.checkpoint = mock_checkpoint
    mock_response.state = state_struct
    
    mock_client._checkpoint_stub.GetCheckpoint = AsyncMock(return_value=mock_response)
    
    result = await mock_client.get_checkpoint("cp_123")
    
    assert result["id"] == "cp_123"
    assert result["execution_id"] == "exec_1"
    assert mock_client._checkpoint_stub.GetCheckpoint.called


@pytest.mark.asyncio
async def test_client_list_checkpoints(mock_client):
    """Test client checkpoint listing."""
    from google.protobuf import timestamp_pb2
    
    # Mock checkpoints
    mock_cp1 = MagicMock()
    mock_cp1.id = "cp_1"
    mock_cp1.checkpoint_name = "checkpoint_1"
    mock_cp1.created_at = timestamp_pb2.Timestamp()
    
    mock_cp2 = MagicMock()
    mock_cp2.id = "cp_2"
    mock_cp2.checkpoint_name = "checkpoint_2"
    mock_cp2.created_at = timestamp_pb2.Timestamp()
    
    mock_response = MagicMock()
    mock_response.success = True
    mock_response.checkpoints = [mock_cp1, mock_cp2]
    
    mock_client._checkpoint_stub.ListCheckpoints = AsyncMock(return_value=mock_response)
    mock_client.set_execution_context("exec_1")
    
    checkpoints = await mock_client.list_checkpoints()
    
    assert len(checkpoints) == 2
    assert checkpoints[0]["id"] == "cp_1"
    assert checkpoints[1]["id"] == "cp_2"


@pytest.mark.asyncio
async def test_client_rollback(mock_client):
    """Test client rollback functionality."""
    mock_target_cp = MagicMock()
    mock_target_cp.id = "cp_target"
    mock_target_cp.checkpoint_name = "target"
    
    mock_response = MagicMock()
    mock_response.success = True
    mock_response.rollback_id = "rb_123"
    mock_response.target_checkpoint = mock_target_cp
    
    mock_client._checkpoint_stub.RollbackToCheckpoint = AsyncMock(return_value=mock_response)
    mock_client.set_execution_context("exec_1")
    
    result = await mock_client.rollback_to_checkpoint("cp_target")
    
    assert result["rollback_id"] == "rb_123"
    assert result["success"] is True
    assert mock_client._checkpoint_stub.RollbackToCheckpoint.called


def test_state_serialization():
    """Test state serialization utility."""
    from omium.checkpoint import _serialize_state
    
    # Test dict
    assert _serialize_state({"key": "value"}) == {"key": "value"}
    
    # Test list
    assert _serialize_state([1, 2, 3]) == [1, 2, 3]
    
    # Test primitive
    assert _serialize_state(42) == 42
    assert _serialize_state("test") == "test"
    
    # Test object with __dict__
    class TestObj:
        def __init__(self):
            self.value = 42
    
    obj = TestObj()
    serialized = _serialize_state(obj)
    assert serialized == {"value": 42}


if __name__ == "__main__":
    pytest.main([__file__, "-v"])




# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

import policy_pb2 as policy__pb2

GRPC_GENERATED_VERSION = '1.71.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in policy_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class PolicyServiceStub(object):
    """Policy Service - Manages recovery policies
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreatePolicy = channel.unary_unary(
                '/omium.policy.v1.PolicyService/CreatePolicy',
                request_serializer=policy__pb2.CreatePolicyRequest.SerializeToString,
                response_deserializer=policy__pb2.CreatePolicyResponse.FromString,
                _registered_method=True)
        self.GetPolicy = channel.unary_unary(
                '/omium.policy.v1.PolicyService/GetPolicy',
                request_serializer=policy__pb2.GetPolicyRequest.SerializeToString,
                response_deserializer=policy__pb2.GetPolicyResponse.FromString,
                _registered_method=True)
        self.ListPolicies = channel.unary_unary(
                '/omium.policy.v1.PolicyService/ListPolicies',
                request_serializer=policy__pb2.ListPoliciesRequest.SerializeToString,
                response_deserializer=policy__pb2.ListPoliciesResponse.FromString,
                _registered_method=True)
        self.UpdatePolicy = channel.unary_unary(
                '/omium.policy.v1.PolicyService/UpdatePolicy',
                request_serializer=policy__pb2.UpdatePolicyRequest.SerializeToString,
                response_deserializer=policy__pb2.UpdatePolicyResponse.FromString,
                _registered_method=True)
        self.DeletePolicy = channel.unary_unary(
                '/omium.policy.v1.PolicyService/DeletePolicy',
                request_serializer=policy__pb2.DeletePolicyRequest.SerializeToString,
                response_deserializer=policy__pb2.DeletePolicyResponse.FromString,
                _registered_method=True)
        self.EvaluatePolicy = channel.unary_unary(
                '/omium.policy.v1.PolicyService/EvaluatePolicy',
                request_serializer=policy__pb2.EvaluatePolicyRequest.SerializeToString,
                response_deserializer=policy__pb2.EvaluatePolicyResponse.FromString,
                _registered_method=True)


class PolicyServiceServicer(object):
    """Policy Service - Manages recovery policies
    """

    def CreatePolicy(self, request, context):
        """Create a policy
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetPolicy(self, request, context):
        """Get a policy
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListPolicies(self, request, context):
        """List policies
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdatePolicy(self, request, context):
        """Update a policy
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeletePolicy(self, request, context):
        """Delete a policy
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EvaluatePolicy(self, request, context):
        """Evaluate a policy
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_PolicyServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CreatePolicy': grpc.unary_unary_rpc_method_handler(
                    servicer.CreatePolicy,
                    request_deserializer=policy__pb2.CreatePolicyRequest.FromString,
                    response_serializer=policy__pb2.CreatePolicyResponse.SerializeToString,
            ),
            'GetPolicy': grpc.unary_unary_rpc_method_handler(
                    servicer.GetPolicy,
                    request_deserializer=policy__pb2.GetPolicyRequest.FromString,
                    response_serializer=policy__pb2.GetPolicyResponse.SerializeToString,
            ),
            'ListPolicies': grpc.unary_unary_rpc_method_handler(
                    servicer.ListPolicies,
                    request_deserializer=policy__pb2.ListPoliciesRequest.FromString,
                    response_serializer=policy__pb2.ListPoliciesResponse.SerializeToString,
            ),
            'UpdatePolicy': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdatePolicy,
                    request_deserializer=policy__pb2.UpdatePolicyRequest.FromString,
                    response_serializer=policy__pb2.UpdatePolicyResponse.SerializeToString,
            ),
            'DeletePolicy': grpc.unary_unary_rpc_method_handler(
                    servicer.DeletePolicy,
                    request_deserializer=policy__pb2.DeletePolicyRequest.FromString,
                    response_serializer=policy__pb2.DeletePolicyResponse.SerializeToString,
            ),
            'EvaluatePolicy': grpc.unary_unary_rpc_method_handler(
                    servicer.EvaluatePolicy,
                    request_deserializer=policy__pb2.EvaluatePolicyRequest.FromString,
                    response_serializer=policy__pb2.EvaluatePolicyResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'omium.policy.v1.PolicyService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('omium.policy.v1.PolicyService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class PolicyService(object):
    """Policy Service - Manages recovery policies
    """

    @staticmethod
    def CreatePolicy(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/omium.policy.v1.PolicyService/CreatePolicy',
            policy__pb2.CreatePolicyRequest.SerializeToString,
            policy__pb2.CreatePolicyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetPolicy(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/omium.policy.v1.PolicyService/GetPolicy',
            policy__pb2.GetPolicyRequest.SerializeToString,
            policy__pb2.GetPolicyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListPolicies(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/omium.policy.v1.PolicyService/ListPolicies',
            policy__pb2.ListPoliciesRequest.SerializeToString,
            policy__pb2.ListPoliciesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UpdatePolicy(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/omium.policy.v1.PolicyService/UpdatePolicy',
            policy__pb2.UpdatePolicyRequest.SerializeToString,
            policy__pb2.UpdatePolicyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DeletePolicy(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/omium.policy.v1.PolicyService/DeletePolicy',
            policy__pb2.DeletePolicyRequest.SerializeToString,
            policy__pb2.DeletePolicyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def EvaluatePolicy(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/omium.policy.v1.PolicyService/EvaluatePolicy',
            policy__pb2.EvaluatePolicyRequest.SerializeToString,
            policy__pb2.EvaluatePolicyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""
Omium Python SDK

Fault-tolerant operating system for production multi-agent AI systems.

Quick Start:
    import omium
    omium.init(api_key="om_xxx")
    
    # Your LangGraph/CrewAI code runs normally
    # Traces and checkpoints are captured automatically

Manual Tracing:
    @omium.trace("my_function")
    def my_function():
        ...

Callbacks:
    from omium import OmiumCallbackHandler
    handler = OmiumCallbackHandler()
    chain.invoke(input, config={"callbacks": [handler]})
"""

from .agent import agent
from .checkpoint import checkpoint as checkpoint_decorator, Checkpoint
from .consensus import consensus, Consensus
from .client import (
    OmiumClient,
    CheckpointError,
    CheckpointNotFoundError,
    CheckpointValidationError,
    ConnectionError
)
from .remote_client import RemoteOmiumClient
from .config import get_config, ConfigManager, OmiumConfig as LegacyOmiumConfig
from .websocket_client import ExecutionWebSocketClient

# New integrations
from .integrations.core import (
    init,
    configure,
    is_initialized,
    get_current_config,
    OmiumConfig,
)
from .integrations.decorators import trace, checkpoint
from .integrations.callbacks import OmiumCallbackHandler
from .integrations.langgraph import instrument_langgraph, uninstrument_langgraph
from .integrations.crewai import instrument_crewai, uninstrument_crewai

__version__ = "0.3.0"
__all__ = [
    # New integration API (primary)
    "init",
    "configure",
    "is_initialized",
    "get_current_config",
    "OmiumConfig",
    "trace",
    "checkpoint",
    "OmiumCallbackHandler",
    "instrument_langgraph",
    "uninstrument_langgraph",
    "instrument_crewai",
    "uninstrument_crewai",
    
    # Legacy API (still supported)
    "agent",
    "checkpoint_decorator",
    "Checkpoint",
    "consensus",
    "Consensus",
    "OmiumClient",
    "RemoteOmiumClient",
    "CheckpointError",
    "CheckpointNotFoundError",
    "CheckpointValidationError",
    "ConnectionError",
    "get_config",
    "ConfigManager",
    "LegacyOmiumConfig",
    "ExecutionWebSocketClient",
]



"""
Omium Decorators - Manual tracing and checkpointing

This module provides decorators for explicit tracing and checkpointing
when auto-instrumentation is not sufficient or desired.
"""

import functools
import asyncio
import logging
import inspect
from typing import Callable, Optional, Any, TypeVar, Union
import uuid

logger = logging.getLogger("omium.decorators")

F = TypeVar('F', bound=Callable[..., Any])


def trace(
    name: Optional[str] = None,
    span_type: str = "function",
    capture_input: bool = True,
    capture_output: bool = True,
    capture_errors: bool = True,
    **default_attributes
) -> Callable[[F], F]:
    """
    Decorator to trace a function.
    
    Automatically creates a span for the decorated function, capturing
    input arguments, output, errors, and timing.
    
    Args:
        name: Span name. Defaults to function name.
        span_type: Type of span (function, tool, llm, etc.)
        capture_input: Whether to capture function arguments
        capture_output: Whether to capture return value
        capture_errors: Whether to capture exceptions
        **default_attributes: Additional attributes to add to every span
    
    Returns:
        Decorated function
    
    Example:
        ```python
        import omium
        
        @omium.trace("research_step")
        def research(topic: str) -> str:
            # ... research logic ...
            return results
        
        @omium.trace(capture_input=False)  # Don't log sensitive inputs
        def process_secret(secret: str) -> bool:
            return validate(secret)
        ```
    """
    def decorator(func: F) -> F:
        span_name = name or func.__name__
        
        # Check if async
        is_async = asyncio.iscoroutinefunction(func)
        
        @functools.wraps(func)
        def sync_wrapper(*args, **kwargs):
            from omium.integrations.core import get_current_config, is_initialized
            from omium.integrations.tracer import OmiumTracer, get_current_tracer
            
            # Skip if not initialized or tracing disabled
            config = get_current_config()
            if not is_initialized() or not config or not config.auto_trace:
                return func(*args, **kwargs)
            
            # Get or create tracer
            tracer = get_current_tracer()
            should_flush = False
            
            if not tracer:
                tracer = OmiumTracer()
                should_flush = True
            
            # Prepare input data
            input_data = None
            if capture_input:
                input_data = _capture_args(func, args, kwargs)
            
            # Create span
            with tracer.span(
                span_name,
                input=input_data,
                span_type=span_type,
                **default_attributes
            ) as span:
                try:
                    result = func(*args, **kwargs)
                    
                    if capture_output:
                        span.set_output(_safe_repr(result))
                    
                    return result
                    
                except Exception as e:
                    if capture_errors:
                        span.set_error(e)
                    raise
                
                finally:
                    if should_flush:
                        tracer.flush()
        
        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            from omium.integrations.core import get_current_config, is_initialized
            from omium.integrations.tracer import OmiumTracer, get_current_tracer
            
            config = get_current_config()
            if not is_initialized() or not config or not config.auto_trace:
                return await func(*args, **kwargs)
            
            tracer = get_current_tracer()
            should_flush = False
            
            if not tracer:
                tracer = OmiumTracer()
                should_flush = True
            
            input_data = None
            if capture_input:
                input_data = _capture_args(func, args, kwargs)
            
            with tracer.span(
                span_name,
                input=input_data,
                span_type=span_type,
                **default_attributes
            ) as span:
                try:
                    result = await func(*args, **kwargs)
                    
                    if capture_output:
                        span.set_output(_safe_repr(result))
                    
                    return result
                    
                except Exception as e:
                    if capture_errors:
                        span.set_error(e)
                    raise
                
                finally:
                    if should_flush:
                        await tracer.aflush()
        
        if is_async:
            return async_wrapper  # type: ignore
        return sync_wrapper  # type: ignore
    
    return decorator


def checkpoint(
    name: str,
    capture_state: bool = True,
    on_error: str = "skip",  # "skip", "raise", "log"
) -> Callable[[F], F]:
    """
    Decorator to create a checkpoint before executing a function.
    
    Checkpoints capture the current state and can be used to replay
    execution from that point.
    
    Args:
        name: Checkpoint name (descriptive, e.g., "after_research", "before_summarize")
        capture_state: Whether to capture function arguments in checkpoint
        on_error: What to do if checkpoint creation fails:
                  - "skip": Continue execution without checkpoint
                  - "raise": Raise the error
                  - "log": Log warning and continue
    
    Returns:
        Decorated function
    
    Example:
        ```python
        import omium
        
        @omium.checkpoint("before_research")
        def research_step(topic: str) -> Dict:
            # ... expensive research ...
            return results
        
        @omium.checkpoint("after_analysis", capture_state=False)
        def analyze(data: Dict) -> str:
            return analysis
        ```
    """
    def decorator(func: F) -> F:
        is_async = asyncio.iscoroutinefunction(func)
        
        @functools.wraps(func)
        def sync_wrapper(*args, **kwargs):
            from omium.integrations.core import get_current_config, is_initialized
            
            config = get_current_config()
            
            # Create checkpoint if enabled
            if is_initialized() and config and config.auto_checkpoint:
                try:
                    _create_checkpoint_sync(
                        name=name,
                        state=_capture_args(func, args, kwargs) if capture_state else None,
                        config=config
                    )
                except Exception as e:
                    if on_error == "raise":
                        raise
                    elif on_error == "log":
                        logger.warning(f"Failed to create checkpoint '{name}': {e}")
                    # "skip" - do nothing
            
            return func(*args, **kwargs)
        
        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            from omium.integrations.core import get_current_config, is_initialized
            
            config = get_current_config()
            
            if is_initialized() and config and config.auto_checkpoint:
                try:
                    await _create_checkpoint_async(
                        name=name,
                        state=_capture_args(func, args, kwargs) if capture_state else None,
                        config=config
                    )
                except Exception as e:
                    if on_error == "raise":
                        raise
                    elif on_error == "log":
                        logger.warning(f"Failed to create checkpoint '{name}': {e}")
            
            return await func(*args, **kwargs)
        
        if is_async:
            return async_wrapper  # type: ignore
        return sync_wrapper  # type: ignore
    
    return decorator


def _capture_args(func: Callable, args: tuple, kwargs: dict) -> dict:
    """
    Capture function arguments as a dictionary.
    
    Tries to map positional args to their parameter names.
    """
    result = {}
    
    try:
        sig = inspect.signature(func)
        params = list(sig.parameters.keys())
        
        # Map positional args
        for i, arg in enumerate(args):
            if i < len(params):
                param_name = params[i]
                # Skip 'self' and 'cls'
                if param_name not in ('self', 'cls'):
                    result[param_name] = _safe_repr(arg)
        
        # Add keyword args
        for key, value in kwargs.items():
            result[key] = _safe_repr(value)
            
    except Exception:
        # Fallback: just use indices
        for i, arg in enumerate(args):
            result[f"arg_{i}"] = _safe_repr(arg)
        result.update({k: _safe_repr(v) for k, v in kwargs.items()})
    
    return result


def _safe_repr(obj: Any, max_length: int = 500) -> Any:
    """
    Safely represent an object, truncating if necessary.
    """
    try:
        if obj is None:
            return None
        
        if isinstance(obj, (str, int, float, bool)):
            if isinstance(obj, str) and len(obj) > max_length:
                return obj[:max_length] + f"... ({len(obj)} total)"
            return obj
        
        if isinstance(obj, dict):
            return {k: _safe_repr(v, max_length // 2) for k, v in list(obj.items())[:20]}
        
        if isinstance(obj, (list, tuple)):
            items = [_safe_repr(item, max_length // 2) for item in obj[:20]]
            if len(obj) > 20:
                items.append(f"... ({len(obj)} total items)")
            return items
        
        # Try string representation
        s = str(obj)
        if len(s) > max_length:
            return s[:max_length] + f"... ({len(s)} total)"
        return s
        
    except Exception:
        return f"<{type(obj).__name__}>"


def _create_checkpoint_sync(
    name: str,
    state: Optional[dict],
    config: Any
) -> Optional[str]:
    """Create a checkpoint synchronously."""
    from omium.integrations.tracer import get_current_tracer
    
    tracer = get_current_tracer()
    execution_id = tracer.execution_id if tracer else str(uuid.uuid4())
    
    logger.info(f"📍 Checkpoint: {name}")
    
    # For now, add as an event in the current span
    # Full checkpoint implementation would use the gRPC client
    if tracer and tracer.current_span:
        tracer.current_span.add_event(f"checkpoint:{name}", {
            "checkpoint_name": name,
            "has_state": state is not None,
        })
    
    # TODO: Full checkpoint implementation
    # try:
    #     from omium.client import OmiumClient
    #     async with OmiumClient(checkpoint_manager_url=config.checkpoint_manager_url) as client:
    #         await client.create_checkpoint(
    #             checkpoint_name=name,
    #             state=state or {},
    #             execution_id=execution_id,
    #         )
    # except Exception as e:
    #     logger.warning(f"Checkpoint creation failed: {e}")
    
    return None


async def _create_checkpoint_async(
    name: str,
    state: Optional[dict],
    config: Any
) -> Optional[str]:
    """Create a checkpoint asynchronously."""
    from omium.integrations.tracer import get_current_tracer
    
    tracer = get_current_tracer()
    execution_id = tracer.execution_id if tracer else str(uuid.uuid4())
    
    logger.info(f"📍 Checkpoint: {name}")
    
    if tracer and tracer.current_span:
        tracer.current_span.add_event(f"checkpoint:{name}", {
            "checkpoint_name": name,
            "has_state": state is not None,
        })
    
    # TODO: Full async checkpoint implementation
    
    return None



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

import tracing_pb2 as tracing__pb2

GRPC_GENERATED_VERSION = '1.71.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in tracing_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class TracingServiceStub(object):
    """Tracing Service - Manages distributed tracing and replay
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SaveSpan = channel.unary_unary(
                '/omium.tracing.v1.TracingService/SaveSpan',
                request_serializer=tracing__pb2.SaveSpanRequest.SerializeToString,
                response_deserializer=tracing__pb2.SaveSpanResponse.FromString,
                _registered_method=True)
        self.GetTrace = channel.unary_unary(
                '/omium.tracing.v1.TracingService/GetTrace',
                request_serializer=tracing__pb2.GetTraceRequest.SerializeToString,
                response_deserializer=tracing__pb2.GetTraceResponse.FromString,
                _registered_method=True)
        self.ListTraces = channel.unary_unary(
                '/omium.tracing.v1.TracingService/ListTraces',
                request_serializer=tracing__pb2.ListTracesRequest.SerializeToString,
                response_deserializer=tracing__pb2.ListTracesResponse.FromString,
                _registered_method=True)
        self.BuildDependencyGraph = channel.unary_unary(
                '/omium.tracing.v1.TracingService/BuildDependencyGraph',
                request_serializer=tracing__pb2.BuildDependencyGraphRequest.SerializeToString,
                response_deserializer=tracing__pb2.BuildDependencyGraphResponse.FromString,
                _registered_method=True)


class TracingServiceServicer(object):
    """Tracing Service - Manages distributed tracing and replay
    """

    def SaveSpan(self, request, context):
        """Save a trace span
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTrace(self, request, context):
        """Get a complete trace
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListTraces(self, request, context):
        """List traces for an execution
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def BuildDependencyGraph(self, request, context):
        """Build dependency graph
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TracingServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'SaveSpan': grpc.unary_unary_rpc_method_handler(
                    servicer.SaveSpan,
                    request_deserializer=tracing__pb2.SaveSpanRequest.FromString,
                    response_serializer=tracing__pb2.SaveSpanResponse.SerializeToString,
            ),
            'GetTrace': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTrace,
                    request_deserializer=tracing__pb2.GetTraceRequest.FromString,
                    response_serializer=tracing__pb2.GetTraceResponse.SerializeToString,
            ),
            'ListTraces': grpc.unary_unary_rpc_method_handler(
                    servicer.ListTraces,
                    request_deserializer=tracing__pb2.ListTracesRequest.FromString,
                    response_serializer=tracing__pb2.ListTracesResponse.SerializeToString,
            ),
            'BuildDependencyGraph': grpc.unary_unary_rpc_method_handler(
                    servicer.BuildDependencyGraph,
                    request_deserializer=tracing__pb2.BuildDependencyGraphRequest.FromString,
                    response_serializer=tracing__pb2.BuildDependencyGraphResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'omium.tracing.v1.TracingService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('omium.tracing.v1.TracingService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class TracingService(object):
    """Tracing Service - Manages distributed tracing and replay
    """

    @staticmethod
    def SaveSpan(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/omium.tracing.v1.TracingService/SaveSpan',
            tracing__pb2.SaveSpanRequest.SerializeToString,
            tracing__pb2.SaveSpanResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTrace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/omium.tracing.v1.TracingService/GetTrace',
            tracing__pb2.GetTraceRequest.SerializeToString,
            tracing__pb2.GetTraceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListTraces(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/omium.tracing.v1.TracingService/ListTraces',
            tracing__pb2.ListTracesRequest.SerializeToString,
            tracing__pb2.ListTracesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def BuildDependencyGraph(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/omium.tracing.v1.TracingService/BuildDependencyGraph',
            tracing__pb2.BuildDependencyGraphRequest.SerializeToString,
            tracing__pb2.BuildDependencyGraphResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)
